#include "poset.h"
#include <pthread.h>

//-----------------------------------------------------------------------------
// Simulation of nsLock.cpp
//
#include <pthread.h>

class nsLock;
static mozilla::poset<nsLock*> ResourceOrdering;

class nsLock {
public:
  void Lock(nsLock* prev) { 
    // deadlock check
    if (prev && ResourceOrdering.ordered(this, prev))
      throw std::logic_error("deadlock!");
    else if (prev)
      ResourceOrdering.order(prev, this, false);

    // PR_Lock()
    pthread_mutex_lock(mLock);
  }

  void Unlock() { 
    pthread_mutex_unlock(mLock);
  }

  static nsLock* NewLock() {
    pthread_mutex_t* lock = new pthread_mutex_t;
    pthread_mutex_init(lock, 0);
    return new nsLock(lock);
  }

  static void DestroyLock(nsLock*& lock) {
    delete lock;
    lock = 0;
  }
private:
  nsLock(pthread_mutex_t* aLock) 
    : mLock(aLock) 
  { }  
  ~nsLock() { 
    ResourceOrdering.remove(this);
    delete mLock; 
    mLock = 0;
  }
  pthread_mutex_t* mLock;
};

static void
AllocLockRecurseUnlockFree(int i, nsLock* prev)
{
  if (0 == i)
    return;

  nsLock* lock = nsLock::NewLock();
  lock->Lock(prev);
  AllocLockRecurseUnlockFree(i-1, lock);
  lock->Unlock();
  nsLock::DestroyLock(lock);
}

int
main(int argc, char** argv)
{
  AllocLockRecurseUnlockFree(1 << 14, 0);
  return 0;
}
